/**
 * Single Window Alt+Tab Library.
 * Needed to install systemwide Alt+Tab -> Ctrl+Tab conversion.
 * Currently it just converts Alt->Ctrl, in fact.
 * 
 * Copyright 2009, Konstantin Tretyakov
 * http://kt.pri.ee/
 *
 * This software and code is subject to the MIT license.
 */
#include "dll.h"

#define SHARED __attribute__((section ("SHARED"), shared))

HHOOK SHARED HookHandle = NULL;
int   SHARED DllLoadCount = 0;
BOOL  SHARED AltIsDown  = FALSE;
int   SHARED AltShiftSequence = 0;
BOOL  SHARED TemporarilyDisabled   = FALSE;
BOOL  SHARED Enabled = TRUE;
HWND  SHARED ListenerHWND = 0;
UINT  SHARED ListenerMessageId = 0;

HINSTANCE hInstance = NULL;

void GenerateKeyEvent(int vk, BOOL keyUp) {
    KEYBDINPUT kb={0};
    INPUT input={0};
    kb.wVk  = vk;  
    if (keyUp) {
        kb.dwFlags  =  KEYEVENTF_KEYUP;
    }
    input.type  = INPUT_KEYBOARD;
    input.ki  = kb;
    SendInput(1,&input,sizeof(input));
}

// Keyboard hook.
DLLIMPORT LRESULT CALLBACK LowLevelKeyboardHook(int nCode, WPARAM wParam, LPARAM lParam) {
    if (nCode >= 0 && !TemporarilyDisabled) {
        // Track the true state of Alt
        DWORD vk =  ((PKBDLLHOOKSTRUCT)lParam)->vkCode;
        BOOL  isKeyUp = (wParam == WM_SYSKEYUP) || (wParam == WM_KEYUP);
        if (vk == VK_LMENU) AltIsDown = isKeyUp ? 0 : 1;
        
        // Detect Alt+Ctrl
        if ((vk == VK_RCONTROL) && isKeyUp && AltIsDown) {
            Enabled = !Enabled;
            
            // If we just enabled the thing, we need to send "Alt up" event before we can continue
            // (there won't be a possibility to create this event later)
            if (Enabled) {
                TemporarilyDisabled = 1;
                GenerateKeyEvent(VK_LMENU, TRUE);
                TemporarilyDisabled = 0;
            }
            else { // Otherwise, send CTRL up 
                 TemporarilyDisabled = 1;
                 GenerateKeyEvent(VK_LCONTROL, TRUE);
                 GenerateKeyEvent(VK_RCONTROL, TRUE);
                 TemporarilyDisabled = 0;
            }
            if (ListenerHWND != 0) PostMessage(ListenerHWND, ListenerMessageId, 0, 0);
        }
        else if (Enabled) { // Do the main part
            // Translate Alt to Ctrl except when Shift is down (this allows to use Alt+Shift as Shift+Alt)
            if ((vk == VK_LMENU) && !(GetKeyState(VK_LSHIFT) & 0x8000)) {
                GenerateKeyEvent(VK_LCONTROL, isKeyUp);
                return 1;
            }  
            /*
            // Second attempt to do the Alt+Shift hack which doesn't work, unfortunately
            if ((vk == VK_LSHIFT) && AltIsDown) {
                if (!isKeyUp) AltShiftSequence = 1;
                else if (AltShiftSequence == 1) {
                    // User pushed Alt+Shift, we temporarily disable ourselves and resend the proper alt+shift
                    TemporarilyDisabled = 1;
                    GenerateKeyEvent(VK_LMENU, 0);
                    GenerateKeyEvent(VK_LSHIFT, 0);
                    GenerateKeyEvent(VK_LSHIFT, 1);
                    GenerateKeyEvent(VK_LMENU, 1);
                    // TODO: Unfortunately, although this sequence gets to execute at proper time, it has no desired effect.
                    TemporarilyDisabled = 0;
                    AltShiftSequence = 0;
                }
            }
            else AltShiftSequence = 0;
            */
        }
    }
    return CallNextHookEx(HookHandle, nCode, wParam, lParam);
}

// Returns false on failure
DLLIMPORT BOOL SaltTabInstallHook(BOOL install) {
    if (install) { // Install hook
        if (HookHandle != NULL) return 0;
        HookHandle  = SetWindowsHookEx(WH_KEYBOARD_LL, LowLevelKeyboardHook, hInstance, 0);
        if(HookHandle == NULL)
            return 0;
        Enabled = 1;
        AltIsDown = 0;
        AltShiftSequence = 0;
        return 1;
    }
    else { // Remove hook
        if (HookHandle == NULL) return 0;
        UnhookWindowsHookEx(HookHandle);
        HookHandle = NULL;
        return 1;
    }
}

DLLIMPORT BOOL SaltTabHookInstalled() {
    return (HookHandle != NULL);
}

DLLIMPORT BOOL SaltTabEnable(BOOL enable) {
    Enabled = enable;
}

DLLIMPORT BOOL SaltTabEnabled() {
    return Enabled;
}

DLLIMPORT void SaltTabSetListener(HWND hWnd, UINT message) {
    ListenerHWND = hWnd; ListenerMessageId = message;
}

DLLIMPORT int SaltTabDllLoadCount() {
    return DllLoadCount;
}

// ------------------------ Bureaucracy of no interest to us ---------------------- //

BOOL APIENTRY DllMain (HINSTANCE hInst     /* Library instance handle. */ ,
                       DWORD reason        /* Reason this function is being called. */ ,
                       LPVOID reserved     /* Not used. */ ) {
    switch (reason)
    {
      case DLL_PROCESS_ATTACH: DllLoadCount++; break;
      case DLL_PROCESS_DETACH: 
        DllLoadCount--;
        // Remove the hook if no one is using the library anymore.
        if (DllLoadCount == 0 && HookHandle != NULL) SaltTabInstallHook(0);
        break;
      case DLL_THREAD_ATTACH:  break;
      case DLL_THREAD_DETACH:  break;
    }
    hInstance = hInst;
    
    /* Returns TRUE on success, FALSE on failure */
    return TRUE;
}
