/**
 * Single Window Alt+Tab Library.
 * Needed to install systemwide Alt+Tab -> Ctrl+Tab conversion.
 * Currently it just converts Alt->Ctrl, in fact.
 * 
 * Copyright 2009, Konstantin Tretyakov
 * http://kt.pri.ee/
 *
 * This software and code is subject to the MIT license.
 */
#include <windows.h>
#include <shellapi.h>
#include "dll.h"
void SetTrayIcon(int action, char* resource_id); // So that I could put the most important function up front
void SetTrayMessage(char* msg);

// ---------------------------------------------------------------------------- //

void UpdateTray() {
    if (SaltTabEnabled()) {
        SetTrayIcon(NIM_MODIFY, "ALTOFF");
        SetTrayMessage("Single-window Alt+Tab enabled");
    }
    else {
        SetTrayIcon(NIM_MODIFY, "ALTON");
        SetTrayMessage("Single-window Alt+Tab disabled");
    }
}
void Toggle() {
    SaltTabEnable(!SaltTabEnabled());
    UpdateTray();
}

// ---------------------------------------------------------------------------- //
// Win32 bureocracy stuff
HWND hWnd;
HINSTANCE hInstance;
char szClassName[] = "SaltTabClass";
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);
#define WM_TRAYMESSAGE (WM_USER + 42)
#define WM_DLLMESSAGE (WM_USER + 43)
NOTIFYICONDATA iconData;
HMENU hMenu; 


// ---------------------------------------------------------------------------- //

// Create a window to receive messages
HWND MakeWindow(HINSTANCE hInstance) {
   
    // The Window structure
    WNDCLASSEX wincl;
    memset(&wincl, 0, sizeof(WNDCLASSEX));
    wincl.hInstance = hInstance;
    wincl.lpszClassName = szClassName;
    wincl.lpfnWndProc = WndProc;
    wincl.cbSize = sizeof (WNDCLASSEX);
    if (!RegisterClassEx (&wincl)) {
        MessageBox(0, "AAAAaaaaaaaa!!!!111", "Error", 0);
        return 0;
    }
    
    return CreateWindow(
           szClassName, NULL,   // Class name, Window name
           0,                   // Style
           0, 0, 0, 0,          // Pos X, Pos Y, Width, Height
           HWND_DESKTOP,        // Parent
           NULL,                // Menu
           hInstance,           // Instance
           NULL                 // CreateParams
           );
}

// ---------------------------------------------------------------------------- //
// Tray icon operations

void SetTrayIcon(int action, char* resource_id) {
   memset(&iconData, 0, sizeof(NOTIFYICONDATA));
   iconData.cbSize = sizeof(NOTIFYICONDATA);
   iconData.hWnd   = hWnd;
   iconData.uFlags = NIF_ICON | NIF_MESSAGE;
   iconData.uCallbackMessage = WM_TRAYMESSAGE;
   iconData.hIcon  = LoadIcon(hInstance, resource_id);
   Shell_NotifyIcon(action, &iconData);
}

void SetTrayMessage(char* msg) {
   strncpy(iconData.szTip, msg, 64);
   iconData.uFlags = NIF_TIP;
   Shell_NotifyIcon(NIM_MODIFY, &iconData);
}

// ---------------------------------------------------------------------------- //

// Program entry point
int WINAPI WinMain (HINSTANCE hThisInstance, HINSTANCE hPrevInstance, LPSTR lpszArgument, int nCmdShow) {
    hInstance = hThisInstance;
    
    if (SaltTabDllLoadCount() > 1) {
        MessageBox(0, "It seems that one SaltTab is already running. You don't need two of these!", "SaltTab Error", MB_ICONERROR);
        return 1;
    }
    
    // Create window
    hWnd = MakeWindow(hInstance);
    if (hWnd == 0) return 1;
    
    // Create menu
    hMenu = LoadMenu(hInstance, "popupmenu");
    hMenu = GetSubMenu(hMenu, 0);

    // Set tray icon
    SetTrayIcon(NIM_ADD, "ALTOFF");
    SetTrayMessage("Single-window Alt+Tab enabled");
    
    // Enable hook
    SaltTabInstallHook(TRUE);
    if (!SaltTabHookInstalled()) {
        MessageBox(0, "For reasons unexplicable and unknown to us, keypress translation could not be enabled. This prevents SaltTab from being any useful :(", "SaltTab Error", MB_ICONERROR);
        return 1;
    }
    SaltTabSetListener(hWnd, WM_DLLMESSAGE);
    
    // Message loop
    MSG message;
    while (GetMessage (&message, NULL, 0, 0)){
        TranslateMessage(&message);
        DispatchMessage(&message);
    }
    
    // Remove tray icon
    SetTrayIcon(NIM_DELETE, NULL);
    
    // Return the value of PostQuitMessage()
    return message.wParam;
}

// ---------------------------------------------------------------------------- //

// The message processing function
LRESULT CALLBACK WndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
    switch (message) {
        case WM_DESTROY:
            PostQuitMessage(0);
            break;
        case WM_TRAYMESSAGE:
            switch(lParam){
               case WM_LBUTTONDOWN:
                    Toggle();
                    break;
               case WM_RBUTTONDOWN:
                  POINT pt;
                  GetCursorPos(&pt);
                  int result = TrackPopupMenu(hMenu, TPM_RETURNCMD | TPM_RIGHTALIGN | TPM_BOTTOMALIGN | TPM_LEFTBUTTON,
                                 pt.x, pt.y, 0, hWnd, 0);
                  if (result == 1) PostQuitMessage(1);
                  else if (result == 4) Toggle();
                  else if (result == 3) MessageBox(0, 
                    "Single-window Alt+Tab\n"
                    "Copyright 2009, Konstantin Tretyakov\n"
                    "http://kt.pri.ee/\n"
                    "\n"
                    "About\n"
                    "-----\n"
                    "This program translates (Left) Alt keypresses into (Left) Ctrl "
                    "keypresses, thus essentially translating Alt+Tab into Ctrl+Tab. "
                    "This allows  you to use Alt+Tab for switching subwindows rather "
                    "than windows  (e.g. tabs in Firefox), which can be more convenient "
                    "than the traditional Ctrl+Tab shortcut used for this purpose.\n"
                    "\n"
                    "Replacing Alt into Ctrl should be harmless for most of people "
                    "most of the time. One example, however, where this is inconvenient "
                    "is the Alt+Shift keyboard layout switch. To somewhat compensate for "
                    "this drawback the program actually won't turn Alt into Ctrl while "
                    "Shift is down, which means that you can switch layout using Shift+Alt "
                    "(rather than Alt+Shift).\n"
                    "\n"
                    "Usage\n"
                    "-----\n"
                    "Just run the executable. You will see a small circle with a crossed "
                    "'Alt' appear on the task tray. Clicking this circle toggles the "
                    "keypress translation on or off. "
                    "Alternatively, you can toggle the translation by pressing Alt+RightCtrl.\n"
                    "\n"
                    "License\n"
                    "-------\n"
                    "Software and source code is subject to the MIT license."
                , "About SaltTab", MB_ICONINFORMATION);
                  break;
            }
            break;
        case WM_DLLMESSAGE:
            UpdateTray();
            break;
        default:
            return DefWindowProc (hwnd, message, wParam, lParam);
    }
    
    return 0;
}
